@@ -23,12 +23,16 @@ https://github.com/telegramdesktop/tdesktop/blob/master/LEGAL
 #include "main/main_session.h"
 #include "core/application.h"
 #include "base/platform/base_platform_info.h"
+#include "base/power_save_blocker.h"
 #include "ui/platform/ui_platform_utility.h"
+#include "ui/platform/ui_platform_window_title.h"
 #include "ui/widgets/buttons.h"
 #include "ui/wrap/fade_wrap.h"
 #include "ui/widgets/shadow.h"
 #include "ui/text/format_values.h"
 #include "ui/gl/gl_surface.h"
+#include "ui/painter.h"
+#include "ui/ui_utility.h"
 #include "window/window_controller.h"
 #include "styles/style_widgets.h"
 #include "styles/style_window.h"
@@ -45,10 +49,6 @@ namespace {
 constexpr auto kPipLoaderPriority = 2;
 constexpr auto kMsInSecond = 1000;
 
-[[nodiscard]] bool IsWindowControlsOnLeft() {
-	return Platform::IsMac();
-}
-
 [[nodiscard]] QRect ScreenFromPosition(QPoint point) {
 	const auto screen = QGuiApplication::screenAt(point);
 	const auto use = screen ? screen : QGuiApplication::primaryScreen();
@@ -57,6 +57,13 @@ constexpr auto kMsInSecond = 1000;
 		: QRect(0, 0, st::windowDefaultWidth, st::windowDefaultHeight);
 }
 
+[[nodiscard]] QSize MaxAllowedSizeForScreen(QSize screenSize) {
+	// Each side should be less than screen side - 3 * st::pipBorderSkip,
+	// That way it won't try to snap to both opposite sides of the screen.
+	const auto skip = 3 * st::pipBorderSkip;
+	return { screenSize.width() - skip, screenSize.height() - skip };
+}
+
 [[nodiscard]] QPoint ClampToEdges(QRect screen, QRect inner) {
 	const auto skip = st::pipBorderSkip;
 	const auto area = st::pipBorderSnapArea;
@@ -344,15 +351,35 @@ void PipPanel::init() {
 	widget()->setMouseTracking(true);
 	widget()->resize(0, 0);
 	widget()->hide();
-	widget()->createWinId();
 
-	rp()->shownValue(
-	) | rpl::filter([=](bool shown) {
-		return shown;
-	}) | rpl::start_with_next([=] {
+	rpl::merge(
+		rp()->shownValue() | rpl::to_empty,
+		rp()->paintRequest() | rpl::to_empty
+	) | rpl::map([=] {
+		return widget()->windowHandle()
+			&& widget()->windowHandle()->isExposed();
+	}) | rpl::distinct_until_changed(
+	) | rpl::filter(rpl::mappers::_1) | rpl::start_with_next([=] {
 		// Workaround Qt's forced transient parent.
 		Ui::Platform::ClearTransientParent(widget());
 	}, rp()->lifetime());
+
+	rp()->shownValue(
+	) | rpl::filter(rpl::mappers::_1) | rpl::start_with_next([=] {
+		Ui::Platform::SetWindowMargins(widget(), _padding);
+	}, rp()->lifetime());
+
+	rp()->screenValue(
+	) | rpl::skip(1) | rpl::start_with_next([=](not_null<QScreen*> screen) {
+		handleScreenChanged(screen);
+	}, rp()->lifetime());
+
+	if (Platform::IsWayland()) {
+		rp()->sizeValue(
+		) | rpl::skip(1) | rpl::start_with_next([=](QSize size) {
+			handleWaylandResize(size);
+		}, rp()->lifetime());
+	}
 }
 
 not_null<QWidget*> PipPanel::widget() const {
@@ -371,6 +398,7 @@ void PipPanel::setAspectRatio(QSize ratio) {
 	if (_ratio.isEmpty()) {
 		_ratio = QSize(1, 1);
 	}
+	Ui::Platform::DisableSystemWindowResize(widget(), _ratio);
 	if (!widget()->size().isEmpty()) {
 		setPosition(countPosition());
 	}
@@ -416,10 +444,7 @@ rpl::producer<> PipPanel::saveGeometryRequests() const {
 }
 
 QScreen *PipPanel::myScreen() const {
-	if (const auto window = widget()->windowHandle()) {
-		return window->screen();
-	}
-	return nullptr;
+	return widget()->screen();
 }
 
 PipPanel::Position PipPanel::countPosition() const {
@@ -461,25 +486,20 @@ PipPanel::Position PipPanel::countPosition() const {
 }
 
 void PipPanel::setPositionDefault() {
-	const auto widgetScreen = [&](auto &&widget) -> QScreen* {
-		if (auto handle = widget ? widget->windowHandle() : nullptr) {
-			return handle->screen();
-		}
-		return nullptr;
-	};
-	const auto parentScreen = widgetScreen(_parent);
-	const auto myScreen = widgetScreen(widget());
-	if (parentScreen && myScreen && myScreen != parentScreen) {
+	const auto parentScreen = _parent ? _parent->screen() : nullptr;
+	const auto myScreen = widget()->screen();
+	if (parentScreen && myScreen != parentScreen) {
+#if QT_VERSION >= QT_VERSION_CHECK(6, 0, 0)
+		widget()->setScreen(parentScreen);
+#else // Qt >= 6.0.0
 		widget()->windowHandle()->setScreen(parentScreen);
+#endif // Qt < 6.0.0
 	}
-	const auto screen = parentScreen
-		? parentScreen
-		: QGuiApplication::primaryScreen();
 	auto position = Position();
 	position.snapped = RectPart::Top | RectPart::Left;
-	position.screen = screen->geometry();
+	position.screen = parentScreen->geometry();
 	position.geometry = QRect(0, 0, st::pipDefaultSize, st::pipDefaultSize);
-	setPositionOnScreen(position, screen->availableGeometry());
+	setPositionOnScreen(position, parentScreen->availableGeometry());
 }
 
 void PipPanel::setPositionOnScreen(Position position, QRect available) {
@@ -492,10 +512,10 @@ void PipPanel::setPositionOnScreen(Position position, QRect available) {
 		? QSize(max, max * _ratio.height() / _ratio.width())
 		: QSize(max * _ratio.width() / _ratio.height(), max);
 
-	// At least one side should not be greater than half of screen size.
-	const auto byWidth = (scaled.width() * screen.height())
-		> (scaled.height() * screen.width());
-	const auto fit = QSize(screen.width() / 2, screen.height() / 2);
+	// Apply maximum size.
+	const auto fit = MaxAllowedSizeForScreen(screen.size());
+	const auto byWidth = (scaled.width() * fit.height())
+		> (scaled.height() * fit.width());
 	const auto normalized = (byWidth && scaled.width() > fit.width())
 		? QSize(fit.width(), fit.width() * scaled.height() / scaled.width())
 		: (!byWidth && scaled.height() > fit.height())
@@ -514,7 +534,7 @@ void PipPanel::setPositionOnScreen(Position position, QRect available) {
 		std::max(normalized.height(), minimalSize.height()));
 
 	// Apply maximal size.
-	const auto maximalSize = (_ratio.width() > _ratio.height())
+	const auto maximalSize = byWidth
 		? QSize(fit.width(), fit.width() * _ratio.height() / _ratio.width())
 		: QSize(fit.height() * _ratio.width() / _ratio.height(), fit.height());
 
@@ -566,6 +586,51 @@ void PipPanel::setGeometry(QRect geometry) {
 	widget()->setGeometry(geometry);
 }
 
+void PipPanel::handleWaylandResize(QSize size) {
+	if (_inHandleWaylandResize) {
+		return;
+	}
+	_inHandleWaylandResize = true;
+
+	// Apply aspect ratio.
+	const auto max = std::max(size.width(), size.height());
+	const auto scaled = (_ratio.width() > _ratio.height())
+		? QSize(max, max * _ratio.height() / _ratio.width())
+		: QSize(max * _ratio.width() / _ratio.height(), max);
+
+	// Buffer can't be bigger than the configured
+	// (suggested by compositor) size.
+	const auto byWidth = (scaled.width() * size.height())
+		> (scaled.height() * size.width());
+	const auto normalized = (byWidth && scaled.width() > size.width())
+		? QSize(size.width(), size.width() * scaled.height() / scaled.width())
+		: (!byWidth && scaled.height() > size.height())
+		? QSize(
+			size.height() * scaled.width() / scaled.height(),
+			size.height())
+		: scaled;
+
+	widget()->resize(normalized);
+	QResizeEvent e(normalized, size);
+	QCoreApplication::sendEvent(widget()->windowHandle(), &e);
+	_inHandleWaylandResize = false;
+}
+
+void PipPanel::handleScreenChanged(not_null<QScreen*> screen) {
+	const auto screenGeometry = screen->availableGeometry();
+	const auto minimalSize = _ratio.scaled(
+		st::pipMinimalSize,
+		st::pipMinimalSize,
+		Qt::KeepAspectRatioByExpanding);
+	const auto maximalSize = _ratio.scaled(
+		MaxAllowedSizeForScreen(screenGeometry.size()),
+		Qt::KeepAspectRatio);
+	widget()->setMinimumSize(minimalSize);
+	widget()->setMaximumSize(
+		std::max(minimalSize.width(), maximalSize.width()),
+		std::max(minimalSize.height(), maximalSize.height()));
+}
+
 void PipPanel::handleMousePress(QPoint position, Qt::MouseButton button) {
 	if (button != Qt::LeftButton) {
 		return;
@@ -677,6 +742,11 @@ void PipPanel::startSystemDrag() {
 	} else {
 		widget()->windowHandle()->startSystemMove();
 	}
+
+	Ui::SendSynteticMouseEvent(
+		widget().get(),
+		QEvent::MouseButtonRelease,
+		Qt::LeftButton);
 }
 
 void PipPanel::processDrag(QPoint point) {
@@ -696,8 +766,7 @@ void PipPanel::processDrag(QPoint point) {
 		st::pipMinimalSize,
 		Qt::KeepAspectRatioByExpanding);
 	const auto maximalSize = _ratio.scaled(
-		screen.width() / 2,
-		screen.height() / 2,
+		MaxAllowedSizeForScreen(screen.size()),
 		Qt::KeepAspectRatio);
 	const auto geometry = Transformed(
 		_dragStartGeometry,
@@ -715,6 +784,10 @@ void PipPanel::processDrag(QPoint point) {
 	const auto clamped = (dragPart == RectPart::Center)
 		? ClampToEdges(screen, valid)
 		: valid.topLeft();
+	widget()->setMinimumSize(minimalSize);
+	widget()->setMaximumSize(
+		std::max(minimalSize.width(), maximalSize.width()),
+		std::max(minimalSize.height(), maximalSize.height()));
 	if (clamped != valid.topLeft()) {
 		moveAnimated(clamped);
 	} else {
@@ -794,8 +867,7 @@ void PipPanel::updateDecorations() {
 		}
 	});
 	const auto position = countPosition();
-	const auto center = position.geometry.center();
-	const auto use = Ui::Platform::TranslucentWindowsSupported(center);
+	const auto use = Ui::Platform::TranslucentWindowsSupported();
 	const auto full = use ? st::callShadow.extend : style::margins();
 	const auto padding = style::margins(
 		(position.attached & RectPart::Left) ? 0 : full.left(),
@@ -811,6 +883,9 @@ void PipPanel::updateDecorations() {
 	_padding = padding;
 	_useTransparency = use;
 	widget()->setAttribute(Qt::WA_OpaquePaintEvent, !_useTransparency);
+	if (widget()->windowHandle()) {
+		Ui::Platform::SetWindowMargins(widget(), _padding);
+	}
 	setGeometry(newGeometry);
 	update();
 }
@@ -818,20 +893,30 @@ void PipPanel::updateDecorations() {
 Pip::Pip(
 	not_null<Delegate*> delegate,
 	not_null<DocumentData*> data,
-	FullMsgId contextId,
+	Data::FileOrigin origin,
+	not_null<DocumentData*> chosenQuality,
+	HistoryItem *context,
+	VideoQuality quality,
 	std::shared_ptr<Streaming::Document> shared,
 	FnMut<void()> closeAndContinue,
 	FnMut<void()> destroy)
 : _delegate(delegate)
 , _data(data)
-, _contextId(contextId)
-, _instance(std::move(shared), [=] { waitingAnimationCallback(); })
+, _origin(origin)
+, _chosenQuality(chosenQuality)
+, _context(context)
+, _quality(quality)
+, _instance(
+	std::in_place,
+	std::move(shared),
+	[=] { waitingAnimationCallback(); })
 , _panel(
 	_delegate->pipParentWidget(),
 	[=](Ui::GL::Capabilities capabilities) {
 		return chooseRenderer(capabilities);
 	})
 , _playbackProgress(std::make_unique<PlaybackProgress>())
+, _dataMedia(_data->createMediaView())
 , _rotation(data->owner().mediaRotation().get(data))
 , _lastPositiveVolume((Core::App().settings().videoVolume() > 0.)
 	? Core::App().settings().videoVolume()
@@ -846,15 +931,28 @@ Pip::Pip(
 	) | rpl::start_with_next([=] {
 		_destroy();
 	}, _panel.rp()->lifetime());
+
+	if (_context) {
+		_data->owner().itemRemoved(
+		) | rpl::start_with_next([=](not_null<const HistoryItem*> data) {
+			if (_context != data) {
+				_context = nullptr;
+			}
+		}, _panel.rp()->lifetime());
+	}
 }
 
 Pip::~Pip() = default;
 
+std::shared_ptr<Streaming::Document> Pip::shared() const {
+	return _instance->shared();
+}
+
 void Pip::setupPanel() {
 	_panel.init();
 	const auto size = [&] {
-		if (!_instance.info().video.size.isEmpty()) {
-			return _instance.info().video.size;
+		if (!_instance->info().video.size.isEmpty()) {
+			return _instance->info().video.size;
 		}
 		const auto media = _data->activeMediaView();
 		if (media) {
@@ -911,7 +1009,7 @@ void Pip::handleLeave() {
 }
 
 void Pip::handleMouseMove(QPoint position) {
-	const auto weak = Ui::MakeWeak(_panel.widget());
+	const auto weak = base::make_weak(_panel.widget());
 	const auto guard = gsl::finally([&] {
 		if (weak) {
 			_panel.handleMouseMove(position);
@@ -989,7 +1087,7 @@ Pip::OverState Pip::ResolveShownOver(OverState state) {
 }
 
 void Pip::handleMousePress(QPoint position, Qt::MouseButton button) {
-	const auto weak = Ui::MakeWeak(_panel.widget());
+	const auto weak = base::make_weak(_panel.widget());
 	const auto guard = gsl::finally([&] {
 		if (weak) {
 			_panel.handleMousePress(position, button);
@@ -1007,10 +1105,7 @@ void Pip::handleMousePress(QPoint position, Qt::MouseButton button) {
 }
 
 void Pip::handleMouseRelease(QPoint position, Qt::MouseButton button) {
-	Expects(1 && _delegate->pipPlaybackSpeed() >= 0.5
-		&& _delegate->pipPlaybackSpeed() <= 2.); // Debugging strange crash.
-
-	const auto weak = Ui::MakeWeak(_panel.widget());
+	const auto weak = base::make_weak(_panel.widget());
 	const auto guard = gsl::finally([&] {
 		if (weak) {
 			_panel.handleMouseRelease(position, button);
@@ -1021,21 +1116,11 @@ void Pip::handleMouseRelease(QPoint position, Qt::MouseButton button) {
 	}
 	seekUpdate(position);
 
-	Assert(2 && _delegate->pipPlaybackSpeed() >= 0.5
-		&& _delegate->pipPlaybackSpeed() <= 2.); // Debugging strange crash.
-
 	volumeControllerUpdate(position);
 
-	Assert(3 && _delegate->pipPlaybackSpeed() >= 0.5
-		&& _delegate->pipPlaybackSpeed() <= 2.); // Debugging strange crash.
-
 	const auto pressed = base::take(_pressed);
 	if (pressed && *pressed == OverState::Playback) {
 		_panel.setDragDisabled(false);
-
-		Assert(4 && _delegate->pipPlaybackSpeed() >= 0.5
-			&& _delegate->pipPlaybackSpeed() <= 2.); // Debugging strange crash.
-
 		seekFinish(_playbackProgress->value());
 	} else if (pressed && *pressed == OverState::VolumeController) {
 		_panel.setDragDisabled(false);
@@ -1086,8 +1171,8 @@ void Pip::seekProgress(float64 value) {
 		_lastDurationMs);
 	if (_seekPositionMs != positionMs) {
 		_seekPositionMs = positionMs;
-		if (!_instance.player().paused()
-			&& !_instance.player().finished()) {
+		if (!_instance->player().paused()
+			&& !_instance->player().finished()) {
 			_pausedBySeek = true;
 			playbackPauseResume();
 		}
@@ -1096,9 +1181,6 @@ void Pip::seekProgress(float64 value) {
 }
 
 void Pip::seekFinish(float64 value) {
-	Expects(5 && _delegate->pipPlaybackSpeed() >= 0.5
-		&& _delegate->pipPlaybackSpeed() <= 2.); // Debugging strange crash.
-
 	if (!_lastDurationMs) {
 		return;
 	}
@@ -1108,7 +1190,7 @@ void Pip::seekFinish(float64 value) {
 		crl::time(0),
 		_lastDurationMs);
 	_seekPositionMs = -1;
-	_startPaused = !_pausedBySeek && !_instance.player().finished();
+	_startPaused = !_pausedBySeek && !_instance->player().finished();
 	restartAtSeekPosition(positionMs);
 }
 
@@ -1185,7 +1267,8 @@ void Pip::setupButtons() {
 			rect.y(),
 			volumeToggleWidth,
 			volumeToggleHeight);
-		if (!IsWindowControlsOnLeft()) {
+		using Ui::Platform::TitleControlsLayout;
+		if (!TitleControlsLayout::Instance()->current().onLeft()) {
 			_close.area.moveLeft(rect.x()
 				+ rect.width()
 				- (_close.area.x() - rect.x())
@@ -1245,18 +1328,71 @@ void Pip::updatePlayPauseResumeState(const Player::TrackState &state) {
 }
 
 void Pip::setupStreaming() {
-	_instance.setPriority(kPipLoaderPriority);
-	_instance.lockPlayer();
+	_instance->setPriority(kPipLoaderPriority);
+	_instance->lockPlayer();
 
-	_instance.player().updates(
+	_instance->switchQualityRequests(
+	) | rpl::filter([=](int quality) {
+		return !_quality.manual && _quality.height != quality;
+	}) | rpl::start_with_next([=](int quality) {
+		applyVideoQuality({
+			.manual = 0,
+			.height = uint32(quality),
+		});
+	}, _instance->lifetime());
+
+	_instance->player().updates(
 	) | rpl::start_with_next_error([=](Streaming::Update &&update) {
 		handleStreamingUpdate(std::move(update));
 	}, [=](Streaming::Error &&error) {
 		handleStreamingError(std::move(error));
-	}, _instance.lifetime());
+	}, _instance->lifetime());
 	updatePlaybackState();
 }
 
+void Pip::applyVideoQuality(VideoQuality value) {
+	if (_quality == value
+		|| !_dataMedia->canBePlayed(_context)) {
+		return;
+	}
+	const auto resolved = _data->chooseQuality(_context, value);
+	if (_chosenQuality == resolved) {
+		return;
+	}
+	auto instance = Streaming::Instance(
+		resolved,
+		_data,
+		_context,
+		_origin,
+		[=] { waitingAnimationCallback(); });
+	if (!instance.valid()) {
+		return;
+	}
+
+	if (_instance->ready()) {
+		_qualityChangeFrame = currentVideoFrameImage();
+	}
+	if (!_instance->player().active()
+		|| _instance->player().finished()) {
+		_qualityChangeFinished = true;
+	}
+	_startPaused = _qualityChangeFinished || _instance->player().paused();
+
+	_quality = value;
+	Core::App().settings().setVideoQuality(value);
+	Core::App().saveSettingsDelayed();
+	_chosenQuality = resolved;
+	_instance.emplace(std::move(instance));
+	setupStreaming();
+	restartAtSeekPosition(_lastUpdatePosition);
+}
+
+QImage Pip::currentVideoFrameImage() const {
+	return _instance->player().ready()
+		? _instance->player().currentFrameImage()
+		: _instance->info().video.cover;
+}
+
 Ui::GL::ChosenRenderer Pip::chooseRenderer(
 		Ui::GL::Capabilities capabilities) {
 	const auto use = Platform::IsMac()
@@ -1287,20 +1423,20 @@ void Pip::paint(not_null<Renderer*> renderer) const {
 		.fade = controlsShown,
 		.outer = _panel.widget()->size(),
 		.rotation = _rotation,
-		.videoRotation = _instance.info().video.rotation,
+		.videoRotation = _instance->info().video.rotation,
 		.useTransparency = _panel.useTransparency(),
 	};
 	if (canUseVideoFrame()) {
 		renderer->paintTransformedVideoFrame(geometry);
-		_instance.markFrameShown();
+		_instance->markFrameShown();
 	} else {
 		const auto content = staticContent();
 		if (_preparedCoverState == ThumbState::Cover) {
 			geometry.rotation += base::take(geometry.videoRotation);
 		}
-		renderer->paintTransformedStaticContent(staticContent(), geometry);
+		renderer->paintTransformedStaticContent(content, geometry);
 	}
-	if (_instance.waitingShown()) {
+	if (_instance->waitingShown()) {
 		renderer->paintRadialLoading(countRadialRect(), controlsShown);
 	}
 	if (controlsShown > 0) {
@@ -1483,28 +1619,31 @@ void Pip::paintVolumeControllerContent(
 void Pip::handleStreamingUpdate(Streaming::Update &&update) {
 	using namespace Streaming;
 
-	v::match(update.data, [&](Information &update) {
+	v::match(update.data, [&](const Information &update) {
 		_panel.setAspectRatio(
 			FlipSizeByRotation(update.video.size, _rotation));
-	}, [&](const PreloadedVideo &update) {
+		_qualityChangeFrame = QImage();
+	}, [&](PreloadedVideo) {
 		updatePlaybackState();
-	}, [&](const UpdateVideo &update) {
+	}, [&](UpdateVideo update) {
 		_panel.update();
 		Core::App().updateNonIdle();
 		updatePlaybackState();
-	}, [&](const PreloadedAudio &update) {
+		_lastUpdatePosition = update.position;
+	}, [&](PreloadedAudio) {
 		updatePlaybackState();
-	}, [&](const UpdateAudio &update) {
+	}, [&](UpdateAudio) {
 		updatePlaybackState();
-	}, [&](WaitingForData) {
-	}, [&](MutedByOther) {
+	}, [](WaitingForData) {
+	}, [](SpeedEstimate) {
+	}, [](MutedByOther) {
 	}, [&](Finished) {
 		updatePlaybackState();
 	});
 }
 
 void Pip::updatePlaybackState() {
-	const auto state = _instance.player().prepareLegacyState();
+	const auto state = _instance->player().prepareLegacyState();
 	updatePlayPauseResumeState(state);
 	if (state.position == kTimeUnknown
 		|| state.length == kTimeUnknown
@@ -1512,6 +1651,7 @@ void Pip::updatePlaybackState() {
 		return;
 	}
 	_playbackProgress->updateState(state);
+	updatePowerSaveBlocker(state);
 
 	qint64 position = 0;
 	if (Player::IsStoppedAtEnd(state.state)) {
@@ -1529,6 +1669,18 @@ void Pip::updatePlaybackState() {
 	}
 }
 
+void Pip::updatePowerSaveBlocker(const Player::TrackState &state) {
+	const auto block = _data->isVideoFile()
+		&& !IsPausedOrPausing(state.state)
+		&& !IsStoppedOrStopping(state.state);
+	base::UpdatePowerSaveBlocker(
+		_powerSaveBlocker,
+		block,
+		base::PowerSaveBlockType::PreventDisplaySleep,
+		[] { return u"Video playback is active"_q; },
+		[=] { return _panel.widget()->windowHandle(); });
+}
+
 void Pip::updatePlaybackTexts(
 		int64 position,
 		int64 length,
@@ -1556,73 +1708,65 @@ void Pip::handleStreamingError(Streaming::Error &&error) {
 }
 
 void Pip::playbackPauseResume() {
-	if (_instance.player().failed()) {
+	if (_instance->player().failed()) {
 		_panel.widget()->close();
-	} else if (_instance.player().finished()
-		|| !_instance.player().active()) {
+	} else if (_instance->player().finished()
+		|| !_instance->player().active()) {
 		_startPaused = false;
 		restartAtSeekPosition(0);
-	} else if (_instance.player().paused()) {
-		_instance.resume();
+	} else if (_instance->player().paused()) {
+		_instance->resume();
 		updatePlaybackState();
 	} else {
-		_instance.pause();
+		_instance->pause();
 		updatePlaybackState();
 	}
 }
 
 void Pip::restartAtSeekPosition(crl::time position) {
-	Expects(6 && _delegate->pipPlaybackSpeed() >= 0.5
-		&& _delegate->pipPlaybackSpeed() <= 2.); // Debugging strange crash.
+	_lastUpdatePosition = position;
 
-	if (!_instance.info().video.cover.isNull()) {
+	if (!_instance->info().video.cover.isNull()) {
 		_preparedCoverStorage = QImage();
 		_preparedCoverState = ThumbState::Empty;
-		_instance.saveFrameToCover();
+		_instance->saveFrameToCover();
 	}
 
-	Assert(7 && _delegate->pipPlaybackSpeed() >= 0.5
-		&& _delegate->pipPlaybackSpeed() <= 2.); // Debugging strange crash.
-
 	auto options = Streaming::PlaybackOptions();
 	options.position = position;
-	options.audioId = _instance.player().prepareLegacyState().id;
-
-	Assert(8 && _delegate->pipPlaybackSpeed() >= 0.5
-		&& _delegate->pipPlaybackSpeed() <= 2.); // Debugging strange crash.
-
+	options.hwAllowed = Core::App().settings().hardwareAcceleratedVideo();
+	options.audioId = _instance->player().prepareLegacyState().id;
 	options.speed = _delegate->pipPlaybackSpeed();
 
-	Assert(9 && options.speed >= 0.5
-		&& options.speed <= 2.); // Debugging strange crash.
-
-	_instance.play(options);
+	_instance->play(options);
 	if (_startPaused) {
-		_instance.pause();
+		_instance->pause();
 	}
 	_pausedBySeek = false;
 	updatePlaybackState();
 }
 
 bool Pip::canUseVideoFrame() const {
-	return _instance.player().ready()
-		&& !_instance.info().video.cover.isNull();
+	return _instance->player().ready()
+		&& !_instance->info().video.cover.isNull();
 }
 
 QImage Pip::videoFrame(const FrameRequest &request) const {
 	Expects(canUseVideoFrame());
 
-	return _instance.frame(request);
+	return _instance->frame(request);
 }
 
 Streaming::FrameWithInfo Pip::videoFrameWithInfo() const {
 	Expects(canUseVideoFrame());
 
-	return _instance.frameWithInfo();
+	return _instance->frameWithInfo();
 }
 
 QImage Pip::staticContent() const {
-	const auto &cover = _instance.info().video.cover;
+	const auto &cover = !_qualityChangeFrame.isNull()
+		? _qualityChangeFrame
+		: _instance->info().video.cover;
 	const auto media = _data->activeMediaView();
 	const auto use = media
 		? media
@@ -1650,7 +1794,7 @@ QImage Pip::staticContent() const {
 	}
 	_preparedCoverState = state;
 	if (state == ThumbState::Cover) {
-		_preparedCoverStorage = _instance.info().video.cover;
+		_preparedCoverStorage = cover;
 	} else {
 		_preparedCoverStorage = (good
 			? good
@@ -1676,7 +1820,7 @@ void Pip::paintRadialLoadingContent(
 		st::radialLine,
 		st::radialLine,
 		st::radialLine));
-	p.setOpacity(_instance.waitingOpacity());
+	p.setOpacity(_instance->waitingOpacity());
 	p.setPen(Qt::NoPen);
 	p.setBrush(st::radialBg);
 	{
@@ -1686,7 +1830,7 @@ void Pip::paintRadialLoadingContent(
 	p.setOpacity(1.);
 	Ui::InfiniteRadialAnimation::Draw(
 		p,
-		_instance.waitingState(),
+		_instance->waitingState(),
 		arc.topLeft(),
 		arc.size(),
 		_panel.widget()->width(),
